//--------------------------------------------------------------------------//
// xmtnimage21.cc                                                           //
// Evaluation of string expressions                                         //
// Latest revision: 04-08-2008                                              //
// Copyright (C) 2008 by Thomas J. Nelson                                   //
// See xmtnimage.h for Copyright Notice                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h" 

extern Globals     g;
extern Image      *z;
extern int         ci;
char *text = NULL;   // make a global so calculator.y can access it
double **fframe=NULL;
double *fframe_1d=NULL;
int fframe_xsize = 0;
int fframe_ysize = 0;
int fframe_allocated = 0;

//-------------------------------------------------------------------------//
// formula_error                                                           //
//-------------------------------------------------------------------------//
void formula_error(char *error_message, char *s)
{
    char tempstring[1024];
    sprintf(tempstring, "%s \n%s", error_message, s);
    tempstring[strlen(tempstring)-2] = 0; // chop off the 255
    message(tempstring, ERROR);
    g.getout=1; 
}

 
//-------------------------------------------------------------------------//
// eval - entry point into evaluating a string                             //
//-------------------------------------------------------------------------//
double eval(char* s, char *answer)
{   
   double dd=0;
   int datatype;
   char *text2;
   if(strlen(s)==0 || strlen(s)>1023) return 0;
   if(g.getout) return 0;
   text2 = new char[strlen(s)+8];
   swap_decimal_point(text2, s);  // this also copies s into text
   text = text2;
   if(strlen(text))
   {   strcat(text,"\255\0\0");
       yyparse();
       dd = read_variable((char*)"temp", answer, datatype);
   }
   delete[] text2;
   return dd;
}    


//-------------------------------------------------------------------------//
//  remove_leading_space                                                   //
//-------------------------------------------------------------------------//
void remove_leading_space(char *s1, char *s2)
{  
   int k, space=0;
   for(k=0;k<(int)strlen(s1);k++) if(s1[k]==' ') space=k+1; else break;
   strcpy(s2, s1+space);
}


//-------------------------------------------------------------------------//
//  remove_trailing_junk  junk being { } ; [space] or \n                   //
//-------------------------------------------------------------------------//
void remove_trailing_junk(char *s)
{
  if(strlen(s)) for(int k=strlen(s); k>=0; k--) if(strchr("{}; \n", s[k])) s[k]=0; else break;
}


//-------------------------------------------------------------------------//
//  remove_quotes - removes 1st & last char in string. Don't just reset    //
//  the ptr or it will crash when they free the memory.                    //
//-------------------------------------------------------------------------//
void remove_quotes(char *s)
{
  int len = (int)strlen(s);
  for(int k=0; k<len-2; k++) s[k]=s[k+1];
  s[len-2] = 0;
}


//-------------------------------------------------------------------------//
// set_value - s can be set(a=b);, set ( a="c" );, set(a=(func(a)));  etc. //
//-------------------------------------------------------------------------//
void set_value(char *s)
{
  char *s2 = strdup(s);
  remove_outer_parentheses(s2, '(', ')'); 
  strcat(s2,";");
  eval(s2, NULL); 
  free(s2);
}


//-------------------------------------------------------------------------//
// check_value                                                             //
//-------------------------------------------------------------------------//
void check_value(char *s)
{
  remove_outer_parentheses(s, '(', ')'); 
  strcat(s,";");
  evaluate(s, 1); // 1 = put in message box
}


//-------------------------------------------------------------------------//
// evaluate  - check a variable or expression                              //
//-------------------------------------------------------------------------//
void evaluate(char *string, int wantmessage)
{
  char **arg;
  arg = new char*[2];
  arg[0] = new char[1];
  arg[1] = new char[FILENAMELENGTH];
  strcpy(arg[1], string);
  evaluate(1, arg, wantmessage);   
  delete[] arg[1];
  delete[] arg[0];
  delete[] arg;
}

void evaluate(int noofargs, char **arg, int wantmessage)
{   
  int j,k, datatype;
  double dd=0;
  char *tempstring = new char[FILENAMELENGTH];
  char *tempstring2 = new char[FILENAMELENGTH];
  char *answer = new char[FILENAMELENGTH];
  if(noofargs<1) return;
  add_all_variables(ci);
  tempstring[0] = 0;
  strcpy(tempstring, arg[1]);
  for(k=2;k<=noofargs;k++)
  {   strcat(tempstring," "); 
      strcat(tempstring,arg[k]); 
  }
  answer[0] = 0;
  //// Try reading it as a variable name first - must strip off
  //// any parentheses, braces, etc. first
  j = 0;
  for(k=0;k<(int)strlen(tempstring);k++)
     if(isalnum(tempstring[k])){ tempstring2[j++]=tempstring[k]; tempstring2[j]=0; }

  dd = read_variable(tempstring2, answer, datatype);

  //// If not a variable, evaluate it as an expression.
  //// Use original string, since it might be sth. like sin(sdf+(erf(2)))
  if(datatype!=STRING && dd==0.0) 
      dd = eval(tempstring, answer);
  if(wantmessage)
  {   switch(datatype)
      {   case STRING: sprintf(tempstring, "string is %s", answer); break;
          case FLOAT:  // fall thru
          default: sprintf(tempstring, "value is %g", dd);
      }
      message(tempstring);
  }
  delete[] answer;
  delete[] tempstring;
  delete[] tempstring2;
}    


//-------------------------------------------------------------------------//
// evaluate_args - allows variable names in macro arguments                //
//-------------------------------------------------------------------------//
void evaluate_args(int noofargs, char **arg)
{   
  int k, datatype=NONE;
  double dd;
  char *answer = new char[FILENAMELENGTH];
  for(k=0; k<=noofargs; k++)  
  {   dd = read_variable(arg[k], answer, datatype);
      switch(datatype)
      {   case FLOAT: 
          case DOUBLE: sprintf(arg[k], "%g", dd); break;
          case STRING: if(strlen(answer)) strcpy(arg[k], answer); break;
      }
  }
  delete[] answer;
}


//-------------------------------------------------------------------------//
// swap_decimal_point                                                      //
//-------------------------------------------------------------------------//
void swap_decimal_point(char* text, char *s)
{
   int k;
   int len = (int)strlen(s);
   for(k=0; k<len; k++) 
   {
       if(s[k]==g.decimal_point && 
          between(k,1,len-1) &&
          isdigit(s[k-1]) && 
          isdigit(s[k+1])) text[k]='.';
       else text[k] = s[k];
   }
   text[len]=0;
}



//--------------------------------------------------------------------------//
//  erf                                                                     //
//--------------------------------------------------------------------------//
#ifndef HAVE_ERF
double erf(double x)
{
   double s1,s2,s3,s4,s5,y;
   s1 = fabs(x);
   s2 = 1 / (1 + .5 * s1);
   s3 = -.82215223 + s2 * .17087277;
   s3 = 1.48851587 + s2 * s3;
   s3 = -1.13520398 + s2 * s3;
   s3 = .27886807 + s2 * s3;            // Chebyshev fitting for
   s3 = -.18628806 + s2 * s3;           // estimation of
   s3 = s2 * (.09678418 + s2 * s3);     // complementary error function
   s3 = s2 * (1.00002368 + s2 * (.37049196 + s3));
   s5 = (-s1 * s1 - 1.26551223 + s3);
   if(s5 < -77) s5 = -77;
   s4 = s2 * exp(s5);
   if(x < 0) s4 = 2 - s4;
   y = s4;
   return y;
}
#endif

//-------------------------------------------------------------------------//
// mask                                                                    //
//-------------------------------------------------------------------------//
void mask(int noofargs, char **arg)
{
  int j,k;
  if(ci<0 && noofargs==0){ message("Select an image first"); return; }
  g.mask_ino1 = ci;
  if(g.mask_ino2 == 0) g.mask_ino2 = ci;
  static Dialog *dialog;

  if(noofargs==0)
  {
     dialog = new Dialog;
     if(dialog==NULL){ message(g.nomemory); return; }
     strcpy(dialog->title,"Mask  (Combine two images)");
     strcpy(dialog->radio[0][0],"Operation"); 
     strcpy(dialog->radio[0][1],"Mask (1&=2)");             
     strcpy(dialog->radio[0][2],"Inverse Mask (1&=~2)");             
     strcpy(dialog->radio[0][3],"Add (1+=2)");
     strcpy(dialog->radio[0][4],"Subtract (1-=2)");
     strcpy(dialog->radio[0][5],"Multiply (1*=2)");
     strcpy(dialog->radio[0][6],"Paste right half");
     dialog->radiono[0] = 7;
     dialog->radioset[0] = g.mask_mode+1;

     strcpy(dialog->boxes[0],"Data");
     strcpy(dialog->boxes[1],"Image to change");
     strcpy(dialog->boxes[2],"Image for mask");
     strcpy(dialog->boxes[3],"X offset");
     strcpy(dialog->boxes[4],"Y offset");

     dialog->boxtype[0]=LABEL;
     dialog->boxtype[1]=INTCLICKBOX;
     dialog->boxtype[2]=INTCLICKBOX;
     dialog->boxtype[3]=INTCLICKBOX;
     dialog->boxtype[4]=INTCLICKBOX;

     dialog->boxmin[1]=1; dialog->boxmax[1]=g.image_count-1;
     dialog->boxmin[2]=1; dialog->boxmax[2]=g.image_count-1;
     dialog->boxmin[3]=-1000; dialog->boxmax[3]=1000;
     dialog->boxmin[4]=-1000; dialog->boxmax[4]=1000;
      
     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;
     itoa(g.mask_ino1, dialog->answer[1][0], 10);
     itoa(g.mask_ino2, dialog->answer[2][0], 10);
     itoa(g.mask_xoffset, dialog->answer[3][0], 10);
     itoa(g.mask_yoffset, dialog->answer[4][0], 10);   

     dialog->noofradios=1;
     dialog->noofboxes=5;
     dialog->helptopic=0;  
     dialog->want_changecicb = 0;
     dialog->f1 = maskcheck;
     dialog->f2 = null;
     dialog->f3 = null;
     dialog->f4 = maskfinish;
     dialog->f5 = null;
     dialog->f6 = null;
     dialog->f7 = null;
     dialog->f8 = null;
     dialog->f9 = null;
     dialog->width = 0;  // calculate automatically
     dialog->height = 0; // calculate automatically
     dialog->transient = 1;
     dialog->wantraise = 0;
     dialog->radiousexy = 0;
     dialog->boxusexy = 0;
     strcpy(dialog->path,".");
     dialog->message[0]=0;      
     dialog->busy = 0;
     dialogbox(dialog);
  }else
  {  if(noofargs>=1) g.mask_ino1 = image_number(arg[1]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=2) g.mask_ino2 = image_number(arg[2]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=3) g.mask_mode = atoi(arg[3]); 
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=4) g.mask_xoffset = atoi(arg[4]); 
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=5) g.mask_yoffset = atoi(arg[5]); 
     do_masking();
  }
}


//--------------------------------------------------------------------------//
//  maskfinish                                                              //
//--------------------------------------------------------------------------//
void maskfinish(dialoginfo *a)
{   
   a=a;
}


//--------------------------------------------------------------------------//
//  maskcheck                                                               //
//--------------------------------------------------------------------------//
void maskcheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  radio=radio; box=box; boxbutton=boxbutton;
  if(radio != -2) return;   //// User clicked Ok or Enter

  g.mask_mode = a->radioset[0] - 1;
  g.mask_ino1 = atoi(a->answer[1][0]);
  g.mask_ino2 = atoi(a->answer[2][0]);
  g.mask_xoffset = atoi(a->answer[3][0]);
  g.mask_yoffset = atoi(a->answer[4][0]);
  if(g.mask_ino1 == g.mask_ino2) 
      if(message("Destination and mask images are the same.\nContinue?",
         YESNOQUESTION) !=YES) return;
  do_masking();
}

//--------------------------------------------------------------------------//
//  do_masking                                                              //
//--------------------------------------------------------------------------//
void do_masking(void)
{
  enum{ TMASK, TIMASK, TADDMASK, TSUBMASK, TMULMASK, TCOMBINE_HALF};
  int bpp, f,i,i2,i3,i4,j,rr1,gg1,bb1,rr2,gg2,bb2,value1,value2, width;
  int ino1 = g.mask_ino1;
  int ino2 = g.mask_ino2;
  if(z[ino1].bpp != z[ino2].bpp){message("Both images must be\nsame bits/pixel"); return; } 
  switch_palette(ino1);
  drawselectbox(OFF);
  bpp = z[ino1].bpp;  
  f = z[ino1].cf;  

  if (g.mask_mode == TCOMBINE_HALF)
  {    width = z[ino1].xsize;
       for(j=0; j<z[ino1].ysize; j++)
       for(i=width/2; i<width; i++)
       {   if(i>=z[ino2].xsize || j>=z[ino2].ysize) continue;
           value1 = pixelat(z[ino2].image[f][j]+i*g.off[bpp], bpp);
           putpixelbytes(z[ino1].image[f][j]+i*g.off[bpp],value1,1,bpp);
       }
  }else if(z[ino1].colortype==GRAY)
  {    for(j=0; j<z[ino1].ysize; j++)
       for(i=0,i2=0,i3=g.mask_xoffset,i4=g.mask_xoffset*g.off[bpp]; i<z[ino1].xsize; 
           i++, i2+=g.off[bpp], i3++, i4+=g.off[bpp])
       {   if(i>=z[ino2].xsize || j>=z[ino2].ysize) continue;
           if(i3>= z[ino1].xsize || j+g.mask_yoffset >= z[ino1].ysize) continue;
           if(i3<0 || j+g.mask_yoffset<0) continue;
           value1 = pixelat(z[ino1].image[f][j+g.mask_yoffset]+i4, bpp);
           value2 = pixelat(z[ino2].image[f][j]+i2, bpp);
           switch(g.mask_mode)
           {   case TMASK: if(!value2) value1=0; break;
               case TIMASK: if(value2) value1=0; break;
               case TADDMASK: value1 += value2; break;
               case TSUBMASK: value1 -= value2; break;
               case TMULMASK: value1 *= value2; break;
           }
           value1 = max(0, min(value1, (int)g.maxvalue[bpp]));
           putpixelbytes(z[ino1].image[f][j+g.mask_yoffset]+i4,value1,1,bpp);
       }
       z[ino1].hitgray = 0;   // Force remapping of gray levels
  }else
  {
       for(j=0; j<z[ino1].ysize; j++)
       for(i=0,i2=0,i3=g.mask_xoffset,i4=g.mask_xoffset*g.off[bpp]; i<z[ino1].xsize; 
           i++, i2+=g.off[bpp], i3++, i4+=g.off[bpp])
       {   if(i>=z[ino2].xsize || j>=z[ino2].ysize) continue;
           if(i3 >= z[ino1].xsize || j+g.mask_yoffset >= z[ino1].ysize) continue;
           if(i3<0 || j+g.mask_yoffset<0) continue;
           value1 = pixelat(z[ino1].image[f][j+g.mask_yoffset]+i4, bpp);
           value2 = pixelat(z[ino2].image[f][j]+i2, bpp);
           valuetoRGB(value1,rr1,gg1,bb1,bpp);
           valuetoRGB(value2,rr2,gg2,bb2,bpp);

           switch(g.mask_mode)
           {   case TMASK: if(!rr2) rr1=0; if(!gg2) gg1=0; if(!bb2) bb1=0; break;
               case TIMASK: if(rr2) rr1=0; if(gg2) gg1=0; if(bb2) bb1=0; break;
               case TADDMASK: rr1+=rr2; gg1+=gg2; bb1+=bb2; break;
               case TSUBMASK: rr1-=rr2; gg1-=gg2; bb1-=bb2; break;
               case TMULMASK: rr1*=rr2; gg1*=gg2; bb1*=bb2; break;
           }
           rr1 = max(0, min(rr1, g.maxred[bpp]));
           gg1 = max(0, min(gg1, g.maxgreen[bpp]));
           bb1 = max(0, min(bb1, g.maxblue[bpp]));
           value1 = RGBvalue(rr1,gg1,bb1,bpp);
           putpixelbytes(z[ino1].image[f][j+g.mask_yoffset]+i4,value1,1,bpp);
       }
  }
  z[ino1].touched=1;
  repair(ino1);
  return;
}



//-------------------------------------------------------------------------//
//  density - coordinates are relative to image.                           //
//-------------------------------------------------------------------------//
double density(int ino, int x1, int x2, int y1, int y2, int compensate, int invert)
{
  int x,y;
  double d = 0;
  for(y=y1-z[ino].ypos; y<=y2-z[ino].ypos; y++)
  for(x=x1-z[ino].xpos; x<=x2-z[ino].xpos; x++)
      d += pixeldensity(x,y,ci,compensate,invert,0);
  return d;
}


//-------------------------------------------------------------------------//
//  area                                                                   //
//-------------------------------------------------------------------------//
double area(int x1, int x2, int y1, int y2)
{
  printf("area %d  %d %d %d\n",x1,y1,x2,y2);
  return 0.0;
}



//-------------------------------------------------------------------------//
// flatframe                                                               //
//-------------------------------------------------------------------------//
void flatframe(int noofargs, char **arg)
{
  int j,k;
  if(ci<0 && noofargs==0){ message("Select an image first"); return; }
  g.flatframe_ino1 = ci;
  if(g.flatframe_ino2 == 0) g.flatframe_ino2 = ci;
  static Dialog *dialog;

  if(noofargs==0)
  {
     dialog = new Dialog;
     if(dialog==NULL){ message(g.nomemory); return; }
     strcpy(dialog->title,"Flat Frame correction");
     strcpy(dialog->radio[0][0],"Operation"); 
     strcpy(dialog->radio[0][1],"Read from file");             
     strcpy(dialog->radio[0][2],"Create from image");             
     strcpy(dialog->radio[0][3],"Save to file");
     strcpy(dialog->radio[0][4],"Apply to image");
     dialog->radiono[0] = 5;
     dialog->radioset[0] = g.flatframe_mode+1;

     strcpy(dialog->boxes[0],"Data");
     strcpy(dialog->boxes[1],"Flatframe file name");
     strcpy(dialog->boxes[2],"Image for flatframe");
     strcpy(dialog->boxes[3],"Image to change");

     dialog->boxtype[0]=LABEL;
     dialog->boxtype[1]=FILENAME;
     dialog->boxtype[2]=INTCLICKBOX;
     dialog->boxtype[3]=INTCLICKBOX;

     strcpy(dialog->answer[1][0], g.flatframe_filename);
     dialog->boxmin[2]=1; dialog->boxmax[2]=g.image_count-1;
     dialog->boxmin[3]=1; dialog->boxmax[3]=g.image_count-1;
      
     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;
     itoa(g.flatframe_ino1, dialog->answer[2][0], 10);
     itoa(g.flatframe_ino2, dialog->answer[3][0], 10);

     dialog->noofradios=1;
     dialog->noofboxes=4;
     dialog->helptopic=0;  
     dialog->want_changecicb = 0;
     dialog->f1 = flatframecheck;
     dialog->f2 = null;
     dialog->f3 = null;
     dialog->f4 = flatframefinish;
     dialog->f5 = null;
     dialog->f6 = null;
     dialog->f7 = null;
     dialog->f8 = null;
     dialog->f9 = null;
     dialog->width = 450;
     dialog->height = 0; // calculate automatically
     dialog->transient = 1;
     dialog->wantraise = 0;
     dialog->radiousexy = 0;
     dialog->boxusexy = 0;
     strcpy(dialog->path,".");
     dialog->message[0]=0;      
     dialog->busy = 0;
     dialogbox(dialog);
  }else
  {  if(noofargs>=1) g.flatframe_ino1 = image_number(arg[1]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=2) g.flatframe_ino2 = image_number(arg[2]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=3) strcpy(g.flatframe_filename, arg[3]);
     if(g.getout){ g.getout=0; return; }
     if(noofargs>=4) g.flatframe_mode = atoi(arg[4]); 
     if(g.getout){ g.getout=0; return; }
     do_flatframe();
  }
}


//--------------------------------------------------------------------------//
//  flatframefinish                                                         //
//--------------------------------------------------------------------------//
void flatframefinish(dialoginfo *a)
{   
   a=a;
}


//--------------------------------------------------------------------------//
//  flatframecheck                                                          //
//--------------------------------------------------------------------------//
void flatframecheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  radio=radio; box=box; boxbutton=boxbutton;
  g.flatframe_mode = a->radioset[0] - 1;
  switch(g.flatframe_mode)
  {
      case 0: 
              setsensitive(a->boxwidget[1][0], True);
              setsensitive(a->boxwidget[2][0], False);
              setsensitive(a->boxwidget[3][0], False);
              break;
      case 1:
              setsensitive(a->boxwidget[1][0], False);
              setsensitive(a->boxwidget[2][0],  True);
              setsensitive(a->boxwidget[3][0], False);
              break;
      case 2:
              setsensitive(a->boxwidget[1][0],  True);
              setsensitive(a->boxwidget[2][0], False);
              setsensitive(a->boxwidget[3][0], False);
              break;
      case 3:
              setsensitive(a->boxwidget[1][0], False);
              setsensitive(a->boxwidget[2][0], False);
              setsensitive(a->boxwidget[3][0],  True);
              break;
  }
  if(radio != -2) return;   //// User clicked Ok or Enter

  strcpy(g.flatframe_filename, a->answer[1][0]);  
  g.flatframe_ino1 = atoi(a->answer[2][0]);
  g.flatframe_ino2 = atoi(a->answer[3][0]);
  do_flatframe();
}


//--------------------------------------------------------------------------//
//  do_flatframe                                                            //
//--------------------------------------------------------------------------//
void do_flatframe(void)
{
  char tempstring[128];
  enum { FLAT_READ, FLAT_CREATE, FLAT_SAVE, FLAT_APPLY };
  int bpp, f,i,j,xsize,ysize,value,errors,errors2;
  int ino1 = g.flatframe_ino1;
  int ino2 = g.flatframe_ino2;
  uchar *add;
  double total, mean;
  switch_palette(ino1);
  drawselectbox(OFF);
  bpp = z[ino1].bpp;  
  f = z[ino1].cf;  
  xsize = z[ino1].xsize;
  ysize = z[ino1].ysize;
  fframe_xsize = xsize;
  fframe_ysize = ysize;

  switch(g.flatframe_mode)
  {   
      case FLAT_READ:
           read_flatframe();
           break;
      case FLAT_CREATE:
           if(z[ino1].colortype != GRAY)
           {   message("Flat frame image must be grayscale", ERROR); 
               return;
           }
           //// calculate mean pixel value           
           total = 0.0;
           for(j=0; j<ysize; j++)
           {    add = z[ino1].image[0][j];
                for(i=0; i<xsize; i++)
                    total += pixelat(add + i*g.off[bpp], bpp);
           }
           mean = total / (double)(xsize * ysize);

           if(flatframe_allocate(xsize, ysize) !=OK) return;

           for(j=0; j<ysize; j++)
           {    add = z[ino1].image[0][j];
                for(i=0; i<xsize; i++)
                {   value = pixelat(add + i*g.off[bpp], bpp);
                    if(!value) value = 1;
                    fframe[j][i] = mean / (double)value;
                }
           }
           sprintf(tempstring, "%d x %d flat frame created", xsize, ysize);
           g.flatframe_xsize = xsize;
           g.flatframe_ysize = ysize;
           message(tempstring);
           break;
      case FLAT_SAVE:
           if(fframe_allocated) save_flatframe();
           else message("Create or read flat frame first");
           break;
      case FLAT_APPLY:
           if(!fframe_allocated) 
           {    message("Create or read flat frame first");
                break;
           }                
           errors = 0;
           errors2 = 0;
           for(j=0; j<ysize; j++)
           {    add = z[ino2].image[0][j];
                for(i=0; i<xsize; i++)
                {   value = pixelat(add + i*g.off[bpp], bpp);
                    if(between(i,0,g.flatframe_xsize-1) &&
                       between(j,0,g.flatframe_ysize-1))
                    {   value = cint((double)value * fframe[j][i]); 
                        if(value <=0 || value >= g.maxvalue[bpp]) errors++;
                        value = max(0, min(value, (int)g.maxvalue[bpp]));
                        putpixelbytes(add+ i*g.off[bpp], value, 1, bpp, 1);
                    }else errors2++;
                }
           }
           z[ino2].touched=1;
           repair(ino2);
           redraw(ino2);
           if(errors)
               sprintf(tempstring, "%d saturated pixels", errors);
           else if(errors2)
               sprintf(tempstring, "Size mismatch for %d x %d flatframe", g.flatframe_xsize,g.flatframe_ysize);
           else
               sprintf(tempstring, "%d x %d Flatframe correction applied", g.flatframe_xsize,g.flatframe_ysize);
           message(tempstring);
           break;
  }

  return;
}


//--------------------------------------------------------------------------//
//  flatframe_allocate                                                      //
//--------------------------------------------------------------------------//
int flatframe_allocate(int xsize, int ysize)
{
   static int oxsize=0, oysize=0;
   if(oxsize || oysize) { free(fframe); free(fframe_1d); }
   fframe_1d = (double*)malloc(xsize * ysize * sizeof(double));
   fframe = make_2d_alias(fframe_1d, xsize, ysize);
   if(fframe==NULL){ message(g.nomemory,ERROR); return NOMEM; }
   oxsize = xsize; 
   oysize = ysize;
   fframe_allocated = 1;
   return OK;
}


//--------------------------------------------------------------------------//
//  read_flatframe                                                          //
//--------------------------------------------------------------------------//
void read_flatframe(void)
{
   int x,y;
   FILE *fp;
   char temp[128];
   if(g.escape || strlen(g.flatframe_filename)==0){ return; }

   if ((fp=fopen(g.flatframe_filename,"r")) == NULL)
   {    error_message(g.flatframe_filename, errno);
   }else
   {   
       fscanf(fp, "%d" ,&g.flatframe_xsize);
       fscanf(fp, "%d" ,&g.flatframe_ysize);
       if(flatframe_allocate(g.flatframe_xsize, g.flatframe_ysize) !=OK) return;
       for(y=0; y<g.flatframe_ysize; y++)
       for(x=0; x<g.flatframe_xsize; x++)
           fscanf(fp, "%lg " ,&fframe[y][x]);
       fclose(fp);
       sprintf(temp,"%d x %d flat frame data read",g.flatframe_xsize,g.flatframe_ysize);
       message(temp); 
   }
}

//--------------------------------------------------------------------------//
//  save_flatframe                                                          //
//--------------------------------------------------------------------------//
void save_flatframe(void)
{
   static PromptStruct ps;
   ps.filename = g.flatframe_filename;
   ps.f1 = save_flatframe_part2;
   ps.f2 = null;
   ps.ptr = NULL;
   ps.param[0] = fframe_xsize;
   ps.param[1] = fframe_ysize;
   check_overwrite_file(&ps);
}


//--------------------------------------------------------------------------//
//  save_flatframe_part2                                                    //
//--------------------------------------------------------------------------//
void save_flatframe_part2(PromptStruct *ps)
{
   char temp[FILENAMELENGTH];
   char *filename = ps->filename;
   int k,x,y;
   int count = ps->n;
   int xsize = ps->param[0];
   int ysize = ps->param[1];

   FILE *fp;
   if ((fp=fopen(filename,"w")) == NULL)
        error_message(filename, errno);
   else
   {   
       fprintf(fp,"%d\n",xsize);
       fprintf(fp,"%d\n",ysize);
       for(y=0; y<ysize; y++)
       for(x=0; x<xsize; x++)
            fprintf(fp,"%lg ", fframe[y][x]);
       for(k=0; k<count; k++)
       temp[0] = 0;
       sprintf(temp,"Flat frame data saved in\n%s",filename);
       message(temp);
       fclose(fp);
   }
}
